

;  yancs - yet another ncurses for scheme
;  strsplit.scm is the scheme program to read the various data files and
;  generate the curslib.c interface source.
;  Copyright (C) 2002 Erik Greenwald <erik@smluc.org>

;  This program is free software; you can redistribute it and/or modify
;  it under the terms of the GNU General Public License as published by
;  the Free Software Foundation; either version 2 of the License, or
;  (at your option) any later version.

;  This program is distributed in the hope that it will be useful,
;  but WITHOUT ANY WARRANTY; without even the implied warranty of
;  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;  GNU General Public License for more details.

;  You should have received a copy of the GNU General Public License
;  along with this program; if not, write to the Free Software
;  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


;(debug-set! stack 0)

(define outconv (list 
    (list "bool"	(lambda (x) (string-append  "SCM_ret = (" x ")==TRUE?SCM_BOOL_T:SCM_BOOL_F")))
    (list "char"	(lambda (x) (string-append  "SCM_ret = gh_char2scm(" x ")")))
    (list "char*"	(lambda (x) (string-append  "SCM_ret = scm_take0str(" x ")")))
    (list "const void*"	(lambda (x) (string-append  "SCM_ret = PTR2SCM(" x ")")))
    (list "int"		(lambda (x) (string-append  "SCM_ret = SCM_MAKINUM(" x ")")))
    (list "unsigned char"	(lambda (x) (string-append  "SCM_ret = gh_char2scm(" x ")")))
    (list "unsigned char*"	(lambda (x) (string-append  "SCM_ret = scm_take0str(" x ")")))
    (list "void"	(lambda (x) (string-append  x ";\n\tSCM_ret = SCM_UNDEFINED")))
    (list "void*"	(lambda (x) (string-append  "SCM_ret = PTR2SCM(" x ")")))
))

(define inconv (list 
    (list "bool"	(lambda (x) (string-append "((" x ")==SCM_BOOL_T)?TRUE:FALSE")))
    (list "char"	(lambda (x) (string-append "SCM_CHAR(" x ")")))
    (list "char*"	(lambda (x) (string-append "SCM_CHARS(" x ")")))
    (list "int"		(lambda (x) (string-append "SCM_INUM(" x ")")))
    (list "short"		(lambda (x) (string-append "(short)SCM_INUM(" x ")")))
    (list "unsigned char"	(lambda (x) (string-append "SCM_CHAR(" x ")")))
    (list "unsigned char*"	(lambda (x) (string-append "SCM_CHARS(" x ")")))
    (list "unsigned int"	(lambda (x) (string-append "(unsigned int)SCM_INUM(" x ")")))
    (list "void*"	(lambda (x) (string-append "SCM2PTR(" x ")")))))

(define (println x) (display x) (newline))
(define (printlnw x) (write x) (newline))

(define (strchr s c) (if (not (and (string? s) (char? c))) -1 (let* ((len (string-length s))) (define (strchr-rec s c v) (if (= v len) -1 (if (char=? c (string-ref s v)) v (strchr-rec s c (+ v 1))))) (strchr-rec s c 0))))

(define (strsplit-char str div)
  (let* ((spl (strchr str div)))
    (if (= -1 spl) (list str) (cons (substring str 0 spl) 
        (strsplit-char (substring str (+ 1 spl) (string-length str)) div)))))

(define (strsplit str div) (cond ((char? div) (strsplit-char str div)); ((string? div) (strsplit-str str div))
    (else (begin (println "strsplit error: type of div") #f))))

(define (cull-const x) (if (and (< 6 (string-length x)) (string=? "const " (substring x 0 6))) (substring x 6 (string-length x)) x))

(define (file-op file mode op) (if (access? file F_OK) (let* ((f (open file mode)) (r (op f))) (close f) r) #f))

(define (file-read file) (define (loadit f) (let* ((x (read-line f))) (if (eof-object? x) '() (if (or (= (string-length x) 0) (char=? (string-ref x 0) #\#)) (loadit f) (cons x (loadit f)))))) (file-op file O_RDONLY loadit))

(define (trim l)
  (define (first-nonwhite l v) (if (char=? (string-ref l v) #\ ) (first-nonwhite l (+ v 1)) v))
  (define (last-nonwhite l v) (if (char=? (string-ref l v) #\ ) (last-nonwhite l (- v 1)) v))
  (substring l (first-nonwhite l 0) (+ (last-nonwhite l (- (string-length l) 1)) 1)))

(define (smobize l)
  (let ((mysmob (strsplit l #\|)))
    (set-cdr! (list-tail outconv (- (length outconv) 1)) (cons (list (car mysmob) (lambda (x) (string-append "SCM_NEWCELL(SCM_ret);\n\tSCM_SETCAR(SCM_ret, SMOB_" 
      (if (char=? (string-ref (car mysmob) (- (string-length (car mysmob)) 1)) #\*) 
        (string-append (substring (car mysmob) 0 (- (string-length (car mysmob)) 1)) "_ptr")
        (car mysmob)) 
      ");\n\tSCM_SETCDR(SCM_ret, (unsigned int)" x ")"))) '()))
    (set-cdr! (list-tail inconv (- (length inconv) 1)) (cons (list (car mysmob) (lambda (x) (string-append "(" (car mysmob) ")SCM_CDR(" x ")"))) '()))
    (set-cdr! (list-tail inconv (- (length inconv) 1)) (cons (list (string-append "const " (car mysmob)) (lambda (x) (string-append "(" x ")"))) '()))
    (list (string-append "SMOB_" 
      (if (char=? (string-ref (car mysmob) (- (string-length (car mysmob)) 1)) #\*) 
        (string-append (substring (car mysmob) 0 (- (string-length (car mysmob)) 1)) "_ptr")
        (car mysmob))) (car mysmob))))

(define (functize l)
  (define (parmsplit l)
    (define (parmsplit-sub l v) (if (or (= v -1) (char=? (string-ref l v) #\ )) v (parmsplit-sub l (- v 1))))
      (let* ((x (parmsplit-sub l (- (string-length l) 1))))
        (if (= x -1) '() (list (substring l 0 x) (string-append "_scm_" (substring l (+ x 1) (string-length l)))))))
  (let* ((funcs (strsplit l #\|))
	 (arglist (map parmsplit (map trim (strsplit (substring (caddr funcs) 1 (- (string-length (caddr funcs)) 1)) #\,)))))
    (list (car funcs) (cadr funcs) (if (null? (car arglist)) '() arglist))))

(define (main args)
  (define (write-def x)
    (define (write-args pre x) (if (null? x) "" (if (null? (cdr x)) (string-append pre (cadar x)) (string-append pre (cadar x) ", " (write-args pre (cdr x))))))
    (display (string-append "SCM\nscm_curses_" (cadr x) "(" (write-args "SCM " (caddr x)) ")\n{\n\tSCM SCM_ret = (SCM)NULL;\n"))
    (map (lambda (y) (display (string-append "\t" (trim (cull-const (car y))) " C_" (cadr y) ";\n"))) (caddr x))
    (map (lambda (y) (display (string-append "\t"
;        "SCM_ASSERT(SCM_NIMP(" (cadr y) ")&&SCM_CAR(" (cadr y) ")==SMOB_"  (if (char=? (string-ref (car y) (- (string-length (car y)) 1)) #\*) (substring (car y) 0 (- (string-length (car y)) 1)) (car y)) ");\n\t"
        "C_" (cadr y) " = " ((cadr (assoc (trim (cull-const (car y))) inconv)) (cadr y)) ";\n"))) 
      (caddr x))
    (display (string-append "\t" ((cadr (assoc (car x) outconv)) (string-append (cadr x) "(" (write-args "C_" (caddr x)) ")")) ";\n\treturn SCM_ret;\n}\n\n")))
  (let* ((funclist (map functize (file-read "funcs")))
         (smoblist (map smobize  (file-read "smobs"))))
    (println "#include <stdio.h>\n#include <stdlib.h>\n#include <curses.h>\n#include <term.h>\n#include <panel.h>\n#include <guile/gh.h>\n#include <libguile.h>\n#include <libguile/smob.h>\n")
    (map (lambda (x) (display "scm_bits_t ") (display (car x)) (display " = (scm_bits_t)0;\n")) smoblist) (newline)
    (map println (file-read "funcs.hand")) (newline)
    (map write-def funclist) (newline)
    (println "void init_curses()\n{\tSCM tmp;\n\n")
    (map (lambda (x) (display (string-append "\t" (car x) " = scm_make_smob_type(\"" (cadr x) "\", sizeof(" (cadr x) "));\n"))) smoblist)
    (map println (file-read "initpre.hand"))
    (newline)
    (map (lambda (x) (display (string-append "\tgh_define (\"" x "\", SCM_MAKINUM(" x "));\n"))) (file-read "defined.ints"))
    (map (lambda (x) (display (string-append "\tSCM_NEWSMOB(tmp, SMOB_" (car x) ", " (cadr x) ");\n" "\tgh_define(\"" (cadr x) "\", tmp);\n"))) (map (lambda (x) (strsplit-char x #\ )) (file-read "smobs.defined")))
    (newline)
    (map (lambda (x) (display (string-append "\tgh_new_procedure(\"" (cadr x) "\", scm_curses_" (cadr x) ", " (number->string (length (caddr x))) ", 0, 0);\n"))) funclist)
    (map println (file-read "initpost.hand"))
    (println "\treturn;\n}\n")))

